/*
**	SampEditClass.c
**
**	Copyright (C) 1995,96 Bernardo Innocenti
**
**	Sample editor class built on top of the "gadgetclass".
*/

#include <string.h>

#include <exec/types.h>
#include <exec/memory.h>
#include <exec/nodes.h>
#include <exec/ports.h>
#include <utility/tagitem.h>
#include <utility/hooks.h>
#include <devices/inputevent.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>
#include <graphics/gfxbase.h>
#include <graphics/gfxmacros.h>
#include <graphics/text.h>
#include <libraries/songclass.h>
#include <libraries/sampeditclass.h>

#include <clib/exec_protos.h>
#include <clib/intuition_protos.h>
#include <clib/graphics_protos.h>
#include <clib/utility_protos.h>
#include <clib/alib_protos.h>

#include <pragmas/exec_sysbase_pragmas.h>
#include <pragmas/intuition_pragmas.h>
#include <pragmas/graphics_pragmas.h>
#include <pragmas/utility_pragmas.h>


/* Some handy definitions missing in <devices/inputevent.h> */
#define IEQUALIFIER_SHIFT	(IEQUALIFIER_LSHIFT | IEQUALIFIER_RSHIFT)
#define IEQUALIFIER_ALT		(IEQUALIFIER_LALT | IEQUALIFIER_RALT)
#define IEQUALIFIER_COMMAND	(IEQUALIFIER_LCOMMAND | IEQUALIFIER_RCOMMAND)

/* Scrolling control */
#define SCROLLED_VERT	1
#define SCROLLED_HORIZ	2



struct SampEditData
{
	struct Instrument	*Instr;

	/* These are the actual Gadget position and size,
	 * regardless of any GREL_#? flags.
	 */
	struct IBox		 GBounds;

	ULONG			 Flags;			/* See definitions in <SampEditClass.h>			*/

	UWORD			 XStart;		/* X offset from the beginning of the sample	*/
	UWORD			 XVisible;		/* Number of visible samples					*/

	UWORD			 CursState;
	UWORD			 CursX;			/* Cursor position (which sample)				*/
	UWORD			 CursScrPosX;	/* Screen cursor position to erase it quickly.	*/

	UWORD			 Pad0;

	UWORD			 BackupCursX;	/* Cursor drag undo */
	UWORD			 BackupXStart;

	ULONG			 XRatio;			/* X Zoom ratio, 16:16 bit fixed point number	*/

	/* Range marking info */
	UWORD			 RangeStartX;
	UWORD			 RangeStartY;
	UWORD			 RangeEndX;
	UWORD			 RangeEndY;

	struct Rectangle RangeRect;		/* Backup of range rect to erase it quickly. */

	/* Rendering routine for the sample. */
	void			(*RenderFunc)(struct Note *note, UBYTE *s);

	/* Pens used to draw various gadget elements */
	ULONG			 BackgroundPen;
	ULONG			 LinePen;
	ULONG			 FillPen;
	ULONG			 GridPen;


	/* For double click test */
	ULONG			 DoubleClickSeconds;
	ULONG			 DoubleClickMicros;
};




/* Function prototypes */

static ULONG __asm SampEditDispatcher (register __a0 Class *cl,
								register __a2 struct ExtGadget *g,
								register __a1 Msg msg);
static ULONG	SAMP_HandleInputMethod	(Class *cl, struct ExtGadget *g, struct gpInput *msg);
static void		SAMP_RenderMethod	(Class *cl, struct ExtGadget *g, struct gpRender *msg);
static ULONG	SAMP_SetMethod		(Class *cl, struct ExtGadget *g, struct opSet *msg);
static ULONG	SAMP_GetMethod		(Class *cl, struct ExtGadget *g, struct opGet *msg);
static ULONG	SAMP_NewMethod		(Class *cl, struct ExtGadget *g, struct opSet *msg);

static void		GetGadgetBox	(struct GadgetInfo *ginfo, struct ExtGadget *g, struct IBox *rect);
static BOOL		CalcDisplaySize	(struct SampEditData *sed, struct ExtGadget *g, struct GadgetInfo *gpi);
static void		SaveUndo		(struct SampEditData *sed);
static BOOL		UndoChange		(struct SampEditData *sed);
static BOOL		RedoChange		(struct SampEditData *sed);
static void		FreeUndoBuffers	(struct SampEditData *sed, BOOL freeall);
static BOOL		MoveCursor		(struct SampEditData *sed, WORD x /*, WORD y*/);
static void		EraseCursor		(struct RastPort *rp, struct SampEditData *sed);
static UWORD	DrawCursor		(struct RastPort *rp, struct SampEditData *sed, struct ExtGadget *g);
static void		DrawRange		(struct RastPort *rp, struct SampEditData *sed);
static void		ClearRange		(struct RastPort *rp, struct SampEditData *sed);
static void		RedrawAll		(struct RastPort *rp, struct SampEditData *sed, struct ExtGadget *g);
static void		RedrawSample	(struct RastPort *rp, struct SampEditData *sed, struct ExtGadget *g);
static void		DrawGrid		(struct RastPort *rp, struct SampEditData *sed, UWORD min, UWORD max);
static UWORD	ScrollDisplay	(struct RastPort *rp, struct SampEditData *sed, struct ExtGadget *g,
								UWORD lefttrack, UWORD topline);
static void		NotifyCursor	(struct ExtGadget *g, struct GadgetInfo *gi, ULONG flags);
static void		NotifyHSlider	(struct ExtGadget *g, struct GadgetInfo *gi, ULONG flags);

//static void __asm TimerIntServer (register __a1 struct Gadget *g);

struct Library * __asm	_UserLibInit	(register __a6 struct Library *mybase);
struct Library * __asm	_UserLibCleanup	(register __a6 struct Library *mybase);
struct IClass * __asm	_GetClass		(register __a6 struct Library *mybase);



/* Library data */

#ifdef _M68020
#define SAMPVERS "_020"
#else
#define SAMPVERS ""
#endif

const UBYTE LibName[] = "sampedit.gadget";
const UBYTE LibVer[] = { '$', 'V', 'E', 'R', ':', ' ' };
const UBYTE LibId[] = "sampedit.gadget" SAMPVERS " 1.1 (5.5.96)  1995,96 by Bernardo Innocenti";




/*****************/
/* Library bases */
/*****************/

/* Get around a SAS/C bug which causes some annoying warnings
 * with the library bases defined below.
 */
#ifdef __SASC
#pragma msg 72 ignore push
#endif /* __SASC */

struct ExecBase			*SysBase		= NULL;
struct IntuitionBase	*IntuitionBase	= NULL;
struct GfxBase			*GfxBase		= NULL;
struct Library			*UtilityBase	= NULL;

#ifdef __SASC
#pragma msg 72 pop
#endif /* __SASC */

static struct IClass	*SampEditClass	= NULL;


static ULONG __asm SampEditDispatcher (register __a0 Class *cl,
										register __a2 struct ExtGadget *g,
										register __a1 Msg msg)

/* SampEdit Class Dispatcher entrypoint.
 * Handle BOOPSI messages.
 */
{
	switch (msg->MethodID)
	{
		case GM_GOACTIVE:
			return SAMP_HandleInputMethod (cl, g, (struct gpInput *)msg);


		case GM_RENDER:
			SAMP_RenderMethod (cl, g, (struct gpRender *)msg);
			return 0;

		case GM_LAYOUT:
		{
			struct SampEditData *sed = INST_DATA (cl, g);

			if (CalcDisplaySize (sed, g, ((struct gpLayout *)msg)->gpl_GInfo))
				NotifyHSlider (g, ((struct gpLayout *)msg)->gpl_GInfo, 0);

			return 0;
		}

		case OM_SET:
		case OM_UPDATE:
			return SAMP_SetMethod (cl, g, (struct opSet *)msg);

		case OM_GET:
			return SAMP_GetMethod (cl, g, (struct opGet *)msg);

		case OM_NEW:
			return (ULONG) SAMP_NewMethod (cl, g, (struct opSet *)msg);

		case OM_DISPOSE:
			/* NOTE: I'm falling through here! */

		default:

			/* Unsupported method: let our superclass's
			 * dispatcher take a look at it.
			 */
			return DoSuperMethodA (cl, (Object *)g, msg);
	}
}



static ULONG SAMP_HandleInputMethod (Class *cl, struct ExtGadget *g, struct gpInput *msg)
{
	struct SampEditData	*sed = INST_DATA (cl, g);
	struct InputEvent	*ie = msg->gpi_IEvent;
	WORD	MouseX, MouseY;		/* Mouse coordinates relative to editing area bounds */
	BOOL	moved		= FALSE,
			scrolled	= FALSE;
	ULONG	result = GMR_MEACTIVE;

	if (msg->MethodID == GM_GOACTIVE)
	{
		if (!sed->Instr)
			return GMR_NOREUSE;

		g->Flags |= GFLG_SELECTED;

		{
			struct RastPort *rp;

			/* Render active cursor */
			if (rp = ObtainGIRPort (((struct gpInput *)msg)->gpi_GInfo))
			{
				DrawCursor (rp, sed, g);
				ReleaseGIRPort (rp);
			}
		}

		/* Do not process InputEvent when the gadget has been
		 * activated by ActivateGadget().
		 */
		if (!((struct gpInput *)msg)->gpi_IEvent)
			return GMR_MEACTIVE;

		/* Note: The input event that triggered the gadget
		 * activation (usually a mouse click) should be passed
		 * to the GM_HANDLEINPUT method, so we fall down to it.
		 */
	}


	MouseX = ((struct gpInput *)msg)->gpi_Mouse.X;
	MouseY = ((struct gpInput *)msg)->gpi_Mouse.Y;

	switch (ie->ie_Class)
	{
		case IECLASS_TIMER:

			/* Timer events are used to keep scrolling
			 * when the mouse is outside the bounds of the
			 * gadget.  When a timer event is recevied and
			 * the mouse button is still pressed, we scroll the
			 * cursor.
			 */
				if (sed->Flags & SEF_DRAGGING)
					moved = MoveCursor (sed, MouseX);

			break;

		case IECLASS_RAWMOUSE:
		{
			BOOL double_click = FALSE;
			BOOL outside = (MouseX < 0) || (MouseX >= sed->GBounds.Width);

			switch (ie->ie_Code)
			{
				case MENUDOWN:

					/* Abort cursor dragging operation and restore
					 * old cursor position.
					 */
					if (sed->Flags & SEF_DRAGGING)
					{
						sed->CursX		= sed->BackupCursX;
						sed->XStart		= sed->BackupXStart;
						moved = TRUE;
						sed->Flags &= ~(SEF_DRAGGING | SEF_SCROLLING);
					}
					else result = GMR_REUSE;

					break;

				case SELECTUP:
					/* Send final update to sliders */
					scrolled = SCROLLED_HORIZ | SCROLLED_VERT;
					sed->Flags &= ~(SEF_DRAGGING | SEF_SCROLLING);
					break;

				case SELECTDOWN:

					/* Check if mouse click is still over the gadget */

					if (outside)
					{
						/* Click outside editing area box:
						 * Check if click is really outside gadget box.
						 * If it is, deactivate the gadget and reuse the event.
						 * Notify application if it is interested in
						 * hearing IDCMP_GADGETUP codes.
						 */

						if	((MouseX < 0) || (MouseX >= sed->GBounds.Width) ||
							(MouseY < 0) || (MouseY >= sed->GBounds.Height))
							result = GMR_REUSE | GMR_VERIFY;
						break;
					}
					else
					{
						/* Backup cursor position for undo feature */
						sed->BackupCursX	= sed->CursX;
						sed->BackupXStart	= sed->XStart;

						/* Start cursor drag mode */
						sed->Flags |= SEF_DRAGGING;
					}

					/* Check for double clicking */

					if (DoubleClick (sed->DoubleClickSeconds, sed->DoubleClickMicros,
						ie->ie_TimeStamp.tv_secs, ie->ie_TimeStamp.tv_micro))
						double_click = TRUE;

					sed->DoubleClickSeconds	= ie->ie_TimeStamp.tv_secs;
					sed->DoubleClickMicros	= ie->ie_TimeStamp.tv_micro;

					/* NOTE: I'm falling through here! */

				default:

					if (sed->Flags & SEF_DRAGGING)
					{
						if (outside)
							sed->Flags |= SEF_SCROLLING;
						else
							moved = MoveCursor (sed, MouseX);
					}
					else if (!moved && double_click)
					{
						if (sed->Flags & SEF_MARKING)
							sed->Flags &= ~SEF_MARKING;
						else
						{
							sed->Flags |= SEF_MARKING;
							sed->RangeStartX = sed->CursX;
						}
/*
						if (rp = ObtainGIRPort (msg->gpi_GInfo))
						{
							if (!(sed->Flags & SEF_MARKING))
								ClearRange (rp, sed);

							DrawCursor (rp, sed, g);

							ReleaseGIRPort (rp);
						}
*/
					}

					break;
			}
			break;
		}

		default:
			return 0;

	}	/* End switch (ie->ie_Class) */

	if (moved)
	{
		struct RastPort *rp;

		if (rp = ObtainGIRPort (msg->gpi_GInfo))
		{
			scrolled |= DrawCursor (rp, sed, g);
			ReleaseGIRPort (rp);
		}

		/* Broadcast notification to our target object. */
		NotifyCursor (g, msg->gpi_GInfo, (ie->ie_Code & IECODE_UP_PREFIX) ? 0 : OPUF_INTERIM);
	}

	if (scrolled)
		NotifyHSlider (g, msg->gpi_GInfo, (ie->ie_Code & IECODE_UP_PREFIX) ? 0 : OPUF_INTERIM);

	return result;
}



static void SAMP_RenderMethod (Class *cl, struct ExtGadget *g, struct gpRender *msg)
{
	struct SampEditData	*sed = INST_DATA (cl, g);

	/* We do not support GREDRAW_UPDATE and GREDRAW_TOGGLE */

	if (msg->gpr_Redraw == GREDRAW_REDRAW)
	{
		/* Recalculate the display size only on V37.
		 * As of V39, Intuition supports GM_LAYOUT, which
		 * allows a more optimized way to handle dynamic resizing.
		 */
		if (IntuitionBase->LibNode.lib_Version < 39)
		{
			if (CalcDisplaySize (sed, g, ((struct gpRender *)msg)->gpr_GInfo))
				NotifyHSlider (g, msg->gpr_GInfo, 0);
		}

		RedrawAll (msg->gpr_RPort, sed, g);
	}
}



static ULONG SAMP_SetMethod (Class *cl, struct ExtGadget *g, struct opSet *msg)
{
	struct SampEditData *sed = INST_DATA (cl, g);
	struct TagItem		*ti,
						*tstate = msg->ops_AttrList;
	BOOL	zoom			= FALSE,
			move_cursor		= FALSE,
			scroll			= FALSE;


	while (ti = NextTagItem(&tstate))
	{
		switch (ti->ti_Tag)
		{
			case SEA_CursXPos:
				if (sed->CursXPos != ti->ti_Data)
				{
					sed->CursXPos = ti->ti_Data;
					move_cursor = TRUE;
				}
				break;

			case SEA_XStart:
				if (sed->XStart != ti->ti_Data)
				{
					sed->XStart = ti->ti_Data;
					scroll = TRUE;
				}
				break;

			case SEA_XRatio:
				if (sed->XRatio != ti->ti_Data)
				{
					sed->XRatio = ti->ti_Data;
					zoom = TRUE;
					break;
				}

			case SEA_ScrollLeft:
				if (sed->XStart > 0)
				{
					sed->XStart--;
					scroll = TRUE;
				}
				break;

			case SEA_ScrollRight:
				if (sed->XStart + sed->XVisible < sed->Lenght)
				{
					sed->XStart++;
					scroll = TRUE;
				}
				break;

			case SEA_MarkRegion:
			{
				struct Rectangle *region = (struct Rectangle *)ti->ti_Data;

				if (!region)				/* End mark mode */

					sed->Flags &= ~SEF_MARKING;

				else if (region == (struct Rectangle *)-1)	/* Toggle mark mode */
				{
					sed->Flags ^= SEF_MARKING;
					sed->RangeStartX	= sed->XStart;
					// sed->RangeStartY	= sed->YStart;
				}
				else						/* Start mark mode */
				{
					memcpy (&sed->RangeStartX, region, sizeof (struct Rectangle));
					sed->CursX	= region->MaxX;
					sed->Flags |= PEF_MARKING;
				}

				if (rp = ObtainGIRPort (((struct opSet *)msg)->ops_GInfo))
				{
					if (!(sed->Flags & PEF_MARKING))
						ClearRange (rp, sed);

					DrawCursor (rp, sed, g);

					ReleaseGIRPort (rp);
				}

				break;
			}

			case SEA_Instrument:
				sed->Instrument = (struct Instrument *) ti->ti_Data;
				redraw = TRUE;
				FreeUndoBuffers (sed, TRUE);

				if (sed->Instrument == NULL)
					sed->CursX = sed->VisibleX = 0;
				else
				{
					/* Recalculate pattern dimensions */
					if (CalcDisplaySize (sed, g, msg->ops_GInfo))
						NotifyHSlider (g, msg->ops_GInfo, 0);
				}
				break;

			case SEA_MaxUndoLevels:
				sed->MaxUndoLevels = ti->ti_Data;
				FreeUndoBuffers (sed, FALSE);
				break;

			case SEA_MaxUndoMem:
				sed->MaxUndoMem = ti->ti_Data;

				/* Unlimited undo memory */
				if (sed->MaxUndoMem == 0) sed->MaxUndoMem = ~0;

				FreeUndoBuffers (sed, FALSE);

				break;

			case SEA_BackgroundPen:
				if (sed->BackgroundPen != ti->ti_Data)
				{
					sed->BackgroundPen = ti->ti_Data;
					redraw_all = TRUE;
				}
				break;

			case SEA_LinePen:
				if (sed->LinePen != ti->ti_Data)
				{
					sed->LinePen = ti->ti_Data;
					redraw_all = TRUE;
				}
				break;

			case SEA_FillPen:
				if (sed->FillPen != ti->ti_Data)
				{
					sed->FillPen = ti->ti_Data;
					redraw_all = TRUE;
				}
				break;

			case SEA_GridPen:
				if (sed->GridPen != ti->ti_Data)
				{
					sed->GridPen = ti->ti_Data;
					redraw_all = TRUE;
				}
				break;

			default:
				break;
		}

	}	/* End while (NextTagItem()) */


	result = DoSuperMethodA (cl, (Object *)g, msg);

	if (redraw_all || scroll || move_cursor || zoom)
	{
		WORD scrolled = 0;

		if (rp = ObtainGIRPort (msg->ops_GInfo))
		{
			if (redraw_all)
			{
				DoMethod ((Object *)g, GM_RENDER, ((struct opSet *)msg)->ops_GInfo, rp, GREDRAW_REDRAW);
				scrolled = 3;
			}
			else if (scroll)
				scrolled = ScrollPattern (rp, sed, g, lefttrack, topline);

			if (move_cursor)
				scrolled |= DrawCursor (rp, sed, g);

			ReleaseGIRPort (rp);
		}

		if (scrolled & 1)
			NotifyVSlider (g, ((struct opSet *)msg)->ops_GInfo, msg->MethodID == OM_UPDATE ? (((struct opUpdate *)msg)->opu_Flags) : 0);
		if (scrolled & 2)
			NotifyHSlider (g, ((struct opSet *)msg)->ops_GInfo, msg->MethodID == OM_UPDATE ? (((struct opUpdate *)msg)->opu_Flags) : 0);

		if (move_cursor || scrolled)
			NotifyCursor (g, ((struct opSet *)msg)->ops_GInfo, msg->MethodID == OM_UPDATE ? (((struct opUpdate *)msg)->opu_Flags) : 0);
	}
}



static LONG SAMP_GetMethod (Class *cl, struct ExtGadget *g, struct opGet *msg)
{
	struct SampEditData *sed = INST_DATA (cl, g);

	switch (msg->opg_AttrID)
	{
		case SEA_CursXPos:
			*msg->opg_Storage = (ULONG) sed->CursXPos;
			break;

		case SEA_CursYPos:
			*msg->opg_Storage = (ULONG) sed->CursYPos;
			break;

		case SEA_XStart:
			*msg->opg_Storage = (ULONG) sed->XStart;
			break;

		case SEA_YStart:
			*msg->opg_Storage = (ULONG) sed->YStart;
			break;

		case SEA_XRatio:
			*msg->opg_Storage = (ULONG) sed->XRatio;
			break;

		case SEA_YRatio:
			*msg->opg_Storage = (ULONG) sed->YRatio;
			break;

		case SEA_MarkRegion:
		{
			struct Rectangle *region = (struct Rectangle *) *msg->opg_Storage);

			region->MinX = min (sed->RangeStartX, sed->RangeEndX);
			region->MaxX = max (sed->RangeStartX, sed->RangeEndX);
			region->MinY = min (sed->RangeStartY, sed->RangeEndY);
			region->MaxY = max (sed->RangeStartY, sed->RangeEndY);
			break;
		}

		case SEA_Flags:
			*msg->opg_Storage = (ULONG) sed->Flags;
			break;

		case SEA_Instrument:
			*msg->opg_Storage = (ULONG) sed->Instrument;
			break;

		default:
			return DoSuperMethodA (cl, (Object *)g, msg);
	}

	return TRUE;
}



static ULONG SAMP_NewMethod (Class *cl, struct ExtGadget *g, struct opSet *msg)
{
	struct SampEditData		*sed;
	ULONG tmp;

	if (!(g = DoSuperMethodA (cl, (Object *)g, msg)))
		return NULL;

	sed = INST_DATA (cl, g);	/* Get pointer to instance data */

	memset (sed, 0, sizeof (struct SampEditData));

	/* We are going to use ScrollRaster() in this gadget... */

	if (g->Flags & GFLG_EXTENDED)
		g->MoreFlags |= GMORE_SCROLLRASTER;

	g->Flags |= GFLG_TABCYCLE | GFLG_RELSPECIAL;

	/* Initialize our lists */
	NEWLIST ((struct List *)&sed->UndoList);
	NEWLIST ((struct List *)&sed->RedoList);


	/* Initialize attributes */

	sed->Instrument		= (struct Instrument *)	GetTagData (SEA_Instrument, NULL, msg->ops_AttrList);
	sed->BackgroundPen	= GetTagData (SEA_BackgroundPen,	0, msg->ops_AttrList);
	sed->LinePen		= GetTagData (SEA_LinePen,			2, msg->ops_AttrList);
	sed->FillPen		= GetTagData (SEA_FillPen,			1, msg->ops_AttrList);
	sed->GridPen		= GetTagData (SEA_GridPen,			3, msg->ops_AttrList);
	sed->MaxUndoLevels	= GetTagData (SEA_MaxUndoLevels,	8, ((struct opSet *)msg)->ops_AttrList);
	sed->MaxUndoMem		= GetTagData (SEA_MaxUndoMem,		32768, ((struct opSet *)msg)->ops_AttrList);
	sed->Flags			= GetTagData (SEA_Flags,			0, ((struct opSet *)msg)->ops_AttrList);

	/* Unlimited undo memory */
	if (sed->MaxUndoMem == 0) sed->MaxUndoMem = ~0;

	return (ULONG) g;
}



static void NotifyCursor (struct ExtGadget *g, struct GadgetInfo *gi, ULONG flags)
{
	/* Always sends notification if the gadget has the GACT_IMMEDIATE
	 * flag set.  If it isn't, the editor will report its cursor
	 * position only the last time it updates the cursor position.
	 */
	if ((g->Activation & GACT_IMMEDIATE) || !(flags & OPUF_INTERIM))
	{
		struct SampEditData *sed = INST_DATA (OCLASS(g), g);

		NotifyAttrChanges ((Object *)g, gi, flags,
			SEA_CursXPos,	sed->CursXPos,
			SEA_CursYPos,	sed->CursYPos,
			GA_ID,			g->GadgetID,
			TAG_DONE);
	}
}



static void NotifyHSlider (struct ExtGadget *g, struct GadgetInfo *gi, ULONG flags)
{
	struct SampEditData *sed = INST_DATA (OCLASS(g), g);

	NotifyAttrChanges ((Object *)g, gi, flags,
		SEA_XStart,	sed->XStart;
		GA_ID,	g->GadgetID,
		TAG_DONE);
}



static void GetGadgetBox (struct GadgetInfo *ginfo, struct ExtGadget *g, struct IBox *rect)

/* This function gets the actual IBox where a gadget exists
 * in a window.  The special cases it handles are all the REL#?
 * (relative positioning flags).
 *
 * The function takes a struct GadgetInfo pointer, a struct Gadget
 * pointer, and a struct IBox pointer.  It uses the window and
 * gadget to fill in the IBox.
 */
{
	rect->Left = g->LeftEdge;
	if (g->Flags & GFLG_RELRIGHT) rect->Left += ginfo->gi_Domain.Width - 1;

	rect->Top = g->TopEdge;
	if (g->Flags & GFLG_RELBOTTOM) rect->Top += ginfo->gi_Domain.Height - 1;

	rect->Width = g->Width;
	if (g->Flags & GFLG_RELWIDTH) rect->Width += ginfo->gi_Domain.Width;

	rect->Height = g->Height;
	if (g->Flags & GFLG_RELHEIGHT) rect->Height += ginfo->gi_Domain.Height;
}



XDEF HOOKCALL struct Library *_UserLibInit (REG(a6,struct Library *mybase))
{
	SysBase = *((struct ExecBase **)4);

	IntuitionBase	= (struct IntuitionBase *) OpenLibrary ("intuition.library", 37);
	GfxBase			= (struct GfxBase *) OpenLibrary ("graphics.library", 37);
	UtilityBase		= OpenLibrary ("utility.library", 37);

	if (!(IntuitionBase && GfxBase && UtilityBase && KeymapBase))
	{
		_UserLibCleanup (mybase);
		return NULL;
	}

	if (SampEditClass = MakeClass (SAMPEDITCLASS, GADGETCLASS, NULL, sizeof (struct SampEditData), 0))
	{
		SampEditClass->cl_Dispatcher.h_Entry = (ULONG (*)()) SampEditDispatcher;
		AddClass (SampEditClass);
	}
	else
	{
		_UserLibCleanup (mybase);
		return NULL;
	}

	return mybase;
}



XDEF HOOKCALL struct Library *_UserLibCleanup (REG(a6,struct Library *mybase))
{
	if (SampEditClass)
		if (!FreeClass (SampEditClass))
			return NULL;

	/* Passing NULL to CloseLibrary() is safe */
	CloseLibrary ((struct Library *)UtilityBase);
	CloseLibrary ((struct Library *)GfxBase);
	CloseLibrary ((struct Library *)IntuitionBase);

	return mybase;
}



XDEF HOOKCALL struct IClass *_GetEngine (REG(a6,struct Library *mybase))
{
	return (SampEditClass);
}
