/*
**	MMDHook.c
**
**	Copyright (C) 1994,95 by Bernardo Innocenti
**
**	External loader and saver hook for MMD0 or MMD1 modules.
**
**	Note: Sorry, this source is a bit of a mess, but the MMD
**	      format is awfully complex <grin>.
**
**
**	Structure of an MMD0/1/2 module as saved by SaveMED():
**
**	MMD0
**	MMD0song
**	MMD0exp
**	InstrExt[]
**	InstrInfo[]
**	<AnnoTxt>
**	<SongName>
**	NotationInfo
**	MMD0blockarr
**	MMD0samparr
**		<blocks>
**		<BlockInfo>		/* Only for MMD1/MMD2 */
**		<BlockName>		/* Only for MMD1/MMD2 */
**	<samples>
*/

#include <exec/types.h>

#include <clib/exec_protos.h>
#include <clib/dos_protos.h>
#include <clib/xmodule_protos.h>

#include <pragmas/exec_sysbase_pragmas.h>
#include <pragmas/dos_pragmas.h>
#include <pragmas/xmodule_pragmas.h>

#include "XModule.h"
#include "Gui.h"
#include "OctaMed.h"


#define OFFSET_MMD0song		sizeof (struct MMD0)
#define OFFSET_MMD0exp		(sizeof (struct MMD0) + sizeof (struct MMD0song))
#define OFFSET_InstrExt		(sizeof (struct MMD0) + sizeof (struct MMD0song) + sizeof (struct MMD0exp))



static INLINE UBYTE DecodeEff (UBYTE eff, UBYTE *effval, UBYTE flags, UBYTE flags2);



/* Effects conversion table */
static const UBYTE Effects[MAXTABLEEFFECTS] =
{
/*   MED		XModule			Val */

 	0x00,	//	Null effect		$00

	0x01,	//	Portamento Up	$01
	0x02,	//	Portamento Down	$02
	0x03,	//	Tone Portamento	$03
	0x14,	//	Vibrato			$04
	0x05,	//	ToneP + VolSl	$05
	0x06,	//	Vibra + VolSl	$06
	0x07,	//	Tremolo			$07
	0x08,	//	Set Hold/Decay	$08
	0x19,	//	Sample Offset	$09
	0x0A,	//	Volume Slide	$0A
	0x0B,	//	Position Jump  	$0B
	0x0C,	//	Set Volume		$0C
	0x1D,	//	Pattern break	$0D
	0x00,	//	Misc			$0E
	0x09,	//	Set Speed		$0F
	0x0F,	//	Set Tempo		$10
	0x00,	//	Arpeggio		$11

	0x00,	//	Oktalyzer H
	0x00,	//	Oktalyzer L
};



LONG GetMED (struct SongInfo *si, BPTR fp)
{
	struct Instrument *inst = si->Inst;
	struct MMD0 mmd0;
	struct MMD0song song;
	ULONG i, j, k, mmdtype, len;

	/******************************/
	/*      Read MMD0 header      */
	/******************************/

	if (Read (fp, &mmd0, sizeof (mmd0)) != sizeof (mmd0))
		return ERR_READWRITE;

	mmdtype = mmd0.id & 0xFF;

	if (GuiSwitches.Verbose)
		ShowMessage (MSG_READING_MMD, mmdtype);

	mmdtype -= '0';

	if (mmdtype > 1)
	{
		ShowMessage (MSG_UNSUPPORTED_MMD_FORMAT);
		return ERR_NOTMODULE;
	}


	/******************************/
	/*  Read MMD0song structure   */
	/******************************/

	if (Seek (fp, (LONG)mmd0.song, OFFSET_BEGINNING) == -1)
		return ERR_READWRITE;

	if (Read (fp, &song, sizeof (song)) != sizeof (song))
		return ERR_READWRITE;


	/* Set instruments parameters */

	for (i = 0; i <= song.numsamples; i++)
	{
		inst[i+1].Volume = song.sample[i].svol;

		if (song.sample[i].rep || song.sample[i].replen != 1)
		{
			inst[i+1].Repeat = song.sample[i].rep << 1;
			inst[i+1].Replen = song.sample[i].replen << 1;
		}
	}

	/* Set position table */
	if (xmSetSongLen (si, song.songlen))
		for (i = 0; i < si->Length; i++)
			si->Sequence[i] = song.playseq[i];


	if (song.flags2 & FLAG2_BPM)
		si->GlobalTempo = (song.deftempo * 4) / ((song.flags2 & FLAG2_BMASK) + 1);
	else
		si->GlobalTempo = song.deftempo;

	si->GlobalSpeed = song.tempo2;



	/******************************/
	/*   Read MMD0exp structure   */
	/******************************/

	if (mmd0.expdata)
	{
		struct MMD0exp exp;

		if (Seek (fp, (LONG)mmd0.expdata, OFFSET_BEGINNING) == -1)
			return ERR_READWRITE;

		if (Read (fp, &exp, sizeof (exp)) != sizeof (exp))
			return ERR_READWRITE;


		/******************************/
		/*  Read InstrExt structures  */
		/******************************/

		if (exp.exp_smp && (exp.s_ext_entrsz >= 4))
		{
			struct InstrExt	instrext;
			ULONG			size = min (exp.s_ext_entrsz, sizeof (instrext));

			if (Seek (fp, (LONG)exp.exp_smp, OFFSET_BEGINNING) == -1)
				return ERR_READWRITE;

			for (i = 1; i <= exp.s_ext_entries; i++)
			{
				if (Read (fp, &instrext, size) != size)
					return ERR_READWRITE;

				inst[i].FineTune = instrext.finetune;
				if (exp.s_ext_entrsz > size)
					if (Seek (fp, exp.s_ext_entrsz - size, OFFSET_CURRENT) == -1)
						return ERR_READWRITE;
			}
		}

		/******************************/
		/*      Read Annotation       */
		/******************************/

		if (exp.annotxt && exp.annolen > 1)
		{
			ULONG len = min (exp.annolen, MAXAUTHNAME-1);

			if (Seek (fp, (LONG)exp.annotxt, OFFSET_BEGINNING) == -1)
				return ERR_READWRITE;

			if (Read (fp, si->Author, len) != len)
				return ERR_READWRITE;
		}


		/******************************/
		/*  Read InstrInfo structures */
		/******************************/

		if (exp.iinfo && (exp.i_ext_entrsz >= sizeof (struct MMDInstrInfo)))
		{
			struct MMDInstrInfo instrinfo;

			if (Seek (fp, (LONG)exp.iinfo, OFFSET_BEGINNING) == -1)
				return ERR_READWRITE;

			for (i = 1; i <= exp.i_ext_entries; i++)
			{
				if (Read (fp, &instrinfo, sizeof (instrinfo)) != sizeof (instrinfo))
					return ERR_READWRITE;

				strncpy (inst[i].Name, instrinfo.name, MAXINSTNAME - 1);

				if (exp.i_ext_entrsz > sizeof (struct MMDInstrInfo))
					if (Seek (fp, exp.i_ext_entrsz - sizeof (struct MMDInstrInfo), OFFSET_CURRENT) == -1)
						return ERR_READWRITE;
			}
		}

		/******************************/
		/*       Read SongName        */
		/******************************/

		if (exp.songname && exp.songnamelen > 1)
		{
			ULONG size = min (exp.songnamelen, MAXSONGNAME-1);

			if (Seek (fp, (LONG)exp.songname, OFFSET_BEGINNING) == -1)
				return ERR_READWRITE;

			if (Read (fp, si->SongName, size) != size)
				return ERR_READWRITE;
		}
	}


	/******************************/
	/*        Read blocarr        */
	/******************************/

	if (mmd0.blockarr)
	{
		LONG *blockarr;

		DisplayAction (MSG_READING_PATTS);

		if (Seek (fp, (LONG)mmd0.blockarr, OFFSET_BEGINNING) == -1)
			return ERR_READWRITE;

		if (!(blockarr = AllocMem (song.numblocks * 4, MEMF_ANY)))
			return ERROR_NO_FREE_STORE;

		if (Read (fp, blockarr, song.numblocks * 4) != song.numblocks * 4)
		{
			FreeMem (blockarr, song.numblocks * 4);
			return ERR_READWRITE;
		}



		/***************/
		/* Read blocks */
		/***************/

		for (i = 0; i < song.numblocks; i++)
		{
			struct Pattern		*patt;
			struct Note			*note;
			struct MMD1block	 block;
			struct MMD0block	 mmd0block;
			struct BlockInfo	 blockinfo;
			ULONG mednote;

			if (DisplayProgress (i, song.numblocks))
			{
				FreeMem (blockarr, song.numblocks * 4);
				return ERROR_BREAK;
			}

			if (Seek (fp, (LONG)blockarr[i], OFFSET_BEGINNING) == -1)
				ShowMessage (MSG_ERR_CANT_LOAD_PATT, i);
			else
			{
				if (mmdtype == 0)
					len = Read (fp, &mmd0block, sizeof (struct MMD0block));
				else
					len = Read (fp, &block, sizeof (struct MMD1block));

				if (len != (mmdtype ? sizeof (struct MMD1block) : sizeof (struct MMD0block)))
					ShowMessage (MSG_ERR_CANT_LOAD_PATT, i);
				else
				{
					if (mmdtype == 0)	/* Convert MMD0 block */
					{
						block.numtracks	= mmd0block.numtracks;
						block.lines	= mmd0block.lines;
						block.info	= 0;
					}

					if (block.numtracks >= MAXTRACKS)
					{
						ShowMessage (MSG_PATT_TOO_MANY_TRACKS, i, MAXTRACKS);
						continue;
					}

					if (si->MaxTracks < block.numtracks)
						si->MaxTracks = block.numtracks;

					if (block.lines > MAXPATTLINES - 1)
					{
						ShowMessage (MSG_PATT_TOO_MANY_LINES, i, MAXPATTLINES);
						block.lines = MAXPATTLINES - 1;
					}

					if (!(patt = AddPattern (si, block.numtracks, block.lines + 1)))
					{
						FreeMem (blockarr, song.numblocks * 4);
						return ERROR_NO_FREE_STORE;
					}

					/***************/
					/* Read Tracks */
					/***************/

					for (k = 0; k < patt->Lines; k++)
					{
						for (j = 0; j < patt->Tracks; j++)
						{
							note = &patt->Notes[j][k];

							if (mmdtype == 0)
							{
								if (Read (fp, &mednote, MMD0ROW) != MMD0ROW)
									return ERR_READWRITE;

								mednote >>= 8;

								note->EffVal = mednote & 0xFF;
								note->EffNum = DecodeEff ((mednote >> 8) & 0x0F, &note->EffVal, song.flags, song.flags2);
								note->Note = ((mednote >> 16) & 0x3F);
								if (note->Note) note->Note += 12;
								note->Inst = ((mednote >> 12) & 0xF) |
									((mednote & (1<<22)) ? 0x20 : 0) | ((mednote & (1<<23)) ? 0x10 : 0);
							}
							else
							{
								if (Read (fp, &mednote, MMD1ROW) != MMD1ROW)
									return ERR_READWRITE;

								note->EffVal = mednote & 0xFF;
								note->EffNum = DecodeEff ((mednote >> 8) & 0xFF, &note->EffVal, song.flags, song.flags2);
								note->Inst = ((mednote >> 16) & 0x3F);
								note->Note = ((mednote >> 24) & 0x7F);

								if (note->Note) note->Note += 12;
							}
						}
					}	/* End for (patt->Tracks) */

					/******************/
					/* Read BlockInfo */
					/******************/

					if (block.info)
						if (Seek (fp, (LONG)block.info, OFFSET_BEGINNING) != -1);
						{
							if (Read (fp, &blockinfo, sizeof (blockinfo)) == sizeof(blockinfo))
							{
								if (blockinfo.blockname)
									if (Seek (fp, (LONG)blockinfo.blockname, OFFSET_BEGINNING) != -1);
									{
										Read (fp, patt->PattName, min (blockinfo.blocknamelen, MAXPATTNAME));
										patt->PattName[MAXPATTNAME-1] = '\0';	/* Ensure NULL-termination */
									}
							}
						}
				}
			}
		}	/* End for (song->numblocks) */

		FreeMem (blockarr, song.numblocks * 4);
	}



	/******************************/
	/*        Read smplarr        */
	/******************************/

	if (mmd0.smplarr)
	{
		LONG *smplarr;
		struct InstrHdr instrhdr;

		DisplayAction (MSG_READING_INSTS);

		if (Seek (fp, (LONG)mmd0.smplarr, OFFSET_BEGINNING) == -1)
			return ERR_READWRITE;

		if (!(smplarr = AllocMem (song.numsamples * 4, MEMF_ANY)))
			return ERROR_NO_FREE_STORE;

		if (Read (fp, smplarr, song.numsamples * 4) != song.numsamples * 4)
		{
			FreeMem (smplarr, song.numsamples * 4);
			return ERR_READWRITE;
		}


		/******************************/
		/*        Read samples        */
		/******************************/

		for (i = 1; i <= song.numsamples ; i++)
		{
			if (!smplarr[i-1]) continue;

			if (DisplayProgress (i, song.numsamples))
			{
				FreeMem (smplarr, song.numsamples * 4);
				return ERROR_BREAK;
			}

			if (Seek (fp, smplarr[i-1], OFFSET_BEGINNING) == -1)
				ShowMessage (MSG_ERR_CANT_LOAD_INST, i);
			else
			{
				if (Read (fp, &instrhdr, sizeof (instrhdr)) != sizeof (instrhdr))
					ShowMessage (MSG_ERR_CANT_LOAD_INST, i);
				else
				{
					switch (instrhdr.type)
					{
						case SAMPLE:
							inst[i].Length = instrhdr.length;

							if (!(inst[i].SampleData = (UBYTE *) AllocVec (inst[i].Length, MEMF_SAMPLE)))
							{
								ShowMessage (MSG_ERR_NO_MEM_FOR_INST, i);
								break;
							}

							if (Read (fp, inst[i].SampleData, inst[i].Length) != inst[i].Length)
								ShowMessage (MSG_ERR_CANT_LOAD_INST, i);

							/* Double sample data */
							if (song.flags & FLAG_8CHANNEL)
							{
								BYTE *data = inst[i].SampleData;

								for (j = 0; j < inst[j].Length; j++)
									*data++ <<= 1;
							}
							break;

						default:
							ShowMessage (MSG_ERR_NOT_A_SAMPLE, i);
							break;
					}
				}
			}
		}

		FreeMem (smplarr, song.numsamples * 4);
	}

	return RETURN_OK;
}



static __inline UBYTE DecodeEff (UBYTE eff, UBYTE *effval, UBYTE flags, UBYTE flags2)
{
	UBYTE i;

	switch (eff)
	{
		case 0x00:
			if (*effval) return EFF_ARPEGGIO;
			else return EFF_NULL;

		case 0x0C:	/* Volume */
			if (!(flags & FLAG_VOLHEX))
				/* Convert decimal volumes */
				*effval = (*effval & 0x0F) + (*effval >> 4) * 10;
			return EFF_SETVOLUME;

		case 0x0F:	/* Tempo */
			if (*effval == 0) return EFF_PATTERNBREAK;
			if (*effval <= 0x0A) return EFF_SETSPEED;
			if (*effval <= 0xF0)
			{
				*effval = (*effval * 4) / ((flags2 & FLAG2_BMASK) + 1);
				return EFF_SETTEMPO;
			}
			if (*effval == 0xF8)
			{
				*effval = 1;
				return EFF_MISC;	/* Filter off */
			}
			if (*effval == 0xF9)
			{
				*effval = 0;
				return EFF_MISC;	/* Filter on */
			}
			return EFF_NULL;

		default:
			for ( i = 0 ; i < MAXTABLEEFFECTS ; i++ )
				if (eff == Effects[i])
					return i;
	}

	return 0;
}



LONG SaveMED (struct SongInfo *si, BPTR fp, UWORD mmd_type)
{
	struct Instrument	*inst = si->Inst;
	struct Pattern		*patt = si->PattData;
	ULONG	modlen, blockslen, instlen, blockptr, offset_blockarr,
			i, j, k, lastinstr;
	BOOL	halve_instr;

	if (GuiSwitches.Verbose)
		ShowMessage (MSG_WRITING_MMD, mmd_type + '0');

	DisplayAction (MSG_WRITING_HEADER);


	/*********************************/
	/* Calculate total patterns size */
	/*********************************/
	{
		if (mmd_type > 0)	/* MMD1 */
		{
			for (i = 0, blockslen = 0; i < si->NumPatterns ; i++)
			{
				blockslen += sizeof (struct MMD1block) +
					(MMD1ROW * patt[i].Lines * patt[i].Tracks) +
					(patt[i].PattName[0] ?
						 (sizeof (struct BlockInfo) + strlen (patt[i].PattName) + 1)
						 : 0);

				if (blockslen & 1) blockslen++; /* Pad to words */
			}
		}
		else	/* MMD0 */
		{
			for (i = 0, blockslen = 0; i < si->NumPatterns ; i++)
			{
				blockslen += sizeof (struct MMD0block) +
					(MMD0ROW * patt[i].Lines * patt[i].Tracks);

				if (blockslen & 1) blockslen++; /* Pad to words */
			}
		}
	}

	/* Find last used instrument */
	for (lastinstr = 63; lastinstr > 0 ; lastinstr--)
		if (inst[lastinstr].Length || inst[lastinstr].Name[0]) break;

	/* Calculate Total instruments size */
	for (i = 1, instlen = 0; i <= lastinstr; i++)
		instlen += (inst[i].Length) ? (si->Inst[i].Length + sizeof (struct InstrHdr)) : 0;

	/* Calculate blockarr offset */
	offset_blockarr = sizeof (struct MMD0) + sizeof (struct MMD0song) + sizeof (struct MMD0exp) +
		lastinstr * (sizeof (struct InstrExt) + sizeof (struct MMDInstrInfo));

	offset_blockarr += strlen (si->Author) + 1;
	if (offset_blockarr & 1) offset_blockarr++;		/* Pad to word */
	offset_blockarr += strlen (si->SongName) + 1;
	if (offset_blockarr & 1) offset_blockarr++;		/* Pad to word */

	/* Calculate Total module size */
	modlen =	offset_blockarr +
				si->NumPatterns * 4 +		/* blockarr		*/
				lastinstr * 4 +				/* samplearr	*/
				blockslen +
				instlen;

	/**********************************/
	/* Fill-in & write module header */
	/*********************************/
	{
		struct MMD0 mmd0;

		memset (&mmd0, 0, sizeof (mmd0));

		mmd0.id			= mmd_type ? ID_MMD1 : ID_MMD0;
		mmd0.modlen 	= modlen;
		mmd0.song		= (struct MMD0song *) OFFSET_MMD0song;
//		mmd0.reserved0	= 0;
		mmd0.blockarr	= (struct MMD0Block **) offset_blockarr;
//		mmd0.reserved1  = si->NumPatterns;
		mmd0.smplarr	= (struct InstrHdr **) (offset_blockarr + (si->NumPatterns * 4));
//		mmd0.reserved2	= lastinst * sizeof (LONG);
		mmd0.expdata	= (struct MMD0exp *) OFFSET_MMD0exp;
//		mmd0.reserved3	= 0;					/* expsize */
//		mmd0.pstate		= 0;					/* the state of the player */
		mmd0.pblock		= si->CurrentPatt;		/* current block */
//		mmd0.pline		= 0;					/* current line */
		mmd0.pseqnum	= si->CurrentPos;		/* current # of playseqlist */
		mmd0.actplayline= -1;					/* OBSOLETE!! DON'T TOUCH! */
//		mmd0.counter	= 0;					/* delay between notes */
//		mmd0.extra_songs= 0;

		/* Write file header */
		if (Write (fp, &mmd0, sizeof (mmd0)) != sizeof (mmd0))
			return ERR_READWRITE;

	}

	/************************************/
	/* Fill-in and write Song structure */
	/************************************/
	{
		struct MMD0song song;

		for (i = 0; i < 63 ; i++)
		{
			song.sample[i].rep			= inst[i+1].Repeat >> 1;
			song.sample[i].replen		= inst[i+1].Replen >> 1;
			song.sample[i].midich		= 0;
			song.sample[i].midipreset	= 0;
			song.sample[i].svol			= inst[i+1].Volume;
			song.sample[i].strans		= 0;
		}

		song.numblocks	= si->NumPatterns;
		song.songlen	= min (si->Length, 256);

		for (i = 0; i < song.songlen; i++)
			song.playseq[i] = si->Sequence[i];

		song.deftempo	= si->GlobalTempo;
		song.playtransp	= 0;
		halve_instr		= si->MaxTracks > 4;
		song.flags		= (halve_instr ? FLAG_8CHANNEL : 0) | FLAG_STSLIDE | FLAG_VOLHEX;
		song.flags2		= 3 | FLAG2_BPM;
		song.tempo2		= si->GlobalSpeed;

		for (i = 0; i <16; i++)
			song.trkvol[i] = 64;

		song.mastervol	= 64;
		song.numsamples	= lastinstr;

		if (Write (fp, &song, sizeof (song)) != sizeof (song))
			return ERR_READWRITE;
	}

	/*****************************/
	/* Write extension (MMD0exp) */
	/*****************************/
	{
		struct MMD0exp mmd0exp;
		LONG len;

		memset (&mmd0exp, 0, sizeof (mmd0exp));

		mmd0exp.exp_smp			= (struct InstrExt *) OFFSET_InstrExt;
		mmd0exp.s_ext_entries	= lastinstr;
		mmd0exp.s_ext_entrsz	= sizeof (struct InstrExt);
		mmd0exp.annotxt			= (STRPTR) (OFFSET_InstrExt + lastinstr * (sizeof (struct InstrExt) + sizeof (struct MMDInstrInfo)));
		mmd0exp.annolen			= strlen (si->Author) + 1;
		mmd0exp.iinfo			= (struct MMDInstrInfo *) (OFFSET_InstrExt + lastinstr * sizeof (struct InstrExt));
		mmd0exp.i_ext_entries	= lastinstr;
		mmd0exp.i_ext_entrsz	= sizeof (struct MMDInstrInfo);
		mmd0exp.songname		= mmd0exp.annotxt + mmd0exp.annolen + ((mmd0exp.annolen & 1) ? 1 : 0);
		mmd0exp.songnamelen		= strlen (si->SongName) + 1;

		if (Write (fp, &mmd0exp, sizeof (mmd0exp)) != sizeof (mmd0exp))
			return ERR_READWRITE;


		/************************/
		/* Write InstrExt array */
		/************************/
		{
			struct InstrExt instrext = { 0 };

			for (i = 1; i <= lastinstr; i++)
			{
				// instrext.hold				= 0;
				// instrext.decay				= 0;
				// instrext.suppress_midi_off	= 0;
				instrext.finetune				= si->Inst[i].FineTune;
				// instrext.default_pitch		= 0;
				// instrext.instr_flags			= 0;
				// instrext.long_midi_preset	= 0;
				// instrext.output_device		= 0;
				// instrext.reserved			= 0;

				if (Write (fp, &instrext, sizeof (instrext)) != sizeof (instrext))
					return ERR_READWRITE;
			}
		}

		/*************************/
		/* Write InstrInfo array */
		/*************************/
		{
			struct MMDInstrInfo instrinfo = { 0 };

			for (i = 1; i <= lastinstr; i++)
			{
				strcpy (instrinfo.name, si->Inst[i].Name);

				if (Write (fp, &instrinfo, sizeof (instrinfo)) != sizeof (instrinfo))
					return ERR_READWRITE;
			}
		}

		/* Write AnnoTxt */

		len = strlen (si->Author) + 1;
		if (len & 1) len++;	/* Pad to WORD */
		if (Write (fp, si->Author, len) != len)
			return ERR_READWRITE;

		/* Write SongName */

		len = strlen (si->SongName) + 1;
		if (len & 1) len++;	/* Pad to WORD */
		if (Write (fp, si->SongName, len) != len)
			return ERR_READWRITE;
	}

	/*************************************/
	/* Write pattern pointers (blockarr) */
	/*************************************/
	{
		blockptr = offset_blockarr + (si->NumPatterns * 4) + (lastinstr * 4);

		for (i = 0; i < si->NumPatterns; i++)
		{
			if (Write (fp, &blockptr, 4) != 4)
				return ERR_READWRITE;

			if (mmd_type)	/* MMD1 and up */
			{
				blockptr += sizeof (struct MMD1block) +
					(MMD1ROW * patt[i].Tracks * patt[i].Lines) +
					((patt[i].PattName[0]) ?
					(sizeof (struct BlockInfo) + strlen (patt[i].PattName) + 1)
					: 0);
			}
			else	/* MMD0 */
			{
				blockptr += sizeof (struct MMD0block) +
					(MMD0ROW * patt[i].Tracks * patt[i].Lines);
			}

			/* Pad to words */
			if (blockptr & 1)	blockptr++;
		}
	}

	/*************************************/
	/* Write sample pointers (samplearr) */
	/*************************************/
	{
		ULONG sampleptr = blockptr, x;

		for (i = 1; i <= lastinstr; i++)
		{
			x = ((inst[i].Length && inst[i].SampleData) ? sampleptr : 0);

			if (Write (fp, &x, 4) != 4)
				return ERR_READWRITE;

			if (x) sampleptr += inst[i].Length + sizeof (struct InstrHdr);
		}
	}

	/********************************/
	/* Write patterns (blocks) data */
	/********************************/
	{
		struct Note	*note;
		ULONG		 mednote;
		LONG		 blockinfoptr = offset_blockarr + (si->NumPatterns * 4) + (lastinstr * 4);
		UBYTE		 eff, effval;
		BOOL		 quiet = FALSE;


		DisplayAction (MSG_WRITING_PATTS);


		for (i = 0; i < si->NumPatterns; i++)
		{
			if (DisplayProgress (i, si->NumPatterns))
				return ERROR_BREAK;

			/* Write Header */

			if (mmd_type)	/* MMD1 and higher */
			{
				struct MMD1block	 block;

				blockinfoptr +=  sizeof (struct MMD1block) +
					(MMD1ROW * patt[i].Tracks * patt[i].Lines);

				block.numtracks = patt[i].Tracks;
				block.lines = patt[i].Lines - 1;

				/* Write BlockInfo only if this pattern has a name */
				block.info = (struct BlockInfo *)(patt[i].PattName[0] ? blockinfoptr : 0);

				if (Write (fp, &block, sizeof (block)) != sizeof (block))
					return ERR_READWRITE;
			}
			else	/* MMD0 */
			{
				struct MMD0block	 block;

				block.numtracks = patt[i].Tracks;
				block.lines = patt[i].Lines - 1;

				if (Write (fp, &block, sizeof (block)) != sizeof (block))
					return ERR_READWRITE;
			}


			/* Write Tracks */
			for (k = 0; k < patt[i].Lines; k++)
			{
				for (j = 0; j < patt[i].Tracks; j++)
				{
					note = &patt[i].Notes[j][k];
					eff = Effects[note->EffNum];

					if (note->EffNum == EFF_MISC)
					{
/*
		pt module	xmod bad conversion	octamed right conversion

		E01			006					0FF8 (Led OFF)
		E11			not converted		1101 (Slide pitch up once)
		EC3			not converted		1803 (Cut note)
		E62			not converted		1602 (Loop)
		601			1601				0601 (Vibrato and fade)
		E93			not converted		1F03 (Note delay and retrigger)
		ED2			????				1F20 (Note delay and retrigger)
		EB1			not converted		1B01 (Slide volume down once)
*/

						switch (note->EffNum >> 4)
						{
							case 0x6:	/* Loop */
								eff = 0x16;
								effval = note->EffNum & 0x0F;
								break;
							case 0xC:	/* Note Cut */
								eff = 0x18;
								break;
							case 0xE:	/* Replay line */
								eff = 0x1E;
								effval = note->EffNum & 0x0F;
								break;
						}
					}
					else if (note->EffNum == EFF_PATTERNBREAK)
						effval = 0;
					else
						effval = note->EffVal;

				//	if (note->EffNum == EFF_SETTEMPO)
				//		effval /= 4;

					if (mmd_type)	/* MMD1 and up */
					{
						mednote = effval |
							((ULONG)eff << 8) |
							((ULONG)(note->Inst & 0x3F) << 16) |
							((ULONG)((note->Note ? (note->Note - 12) : 0) & 0x7F) << 24);

						if (Write (fp, &mednote, MMD1ROW) != MMD1ROW)
							return ERR_READWRITE;
					}
					else
					{
						if ((eff > 0x0F) && !quiet)
						{
							ShowMessage (MSG_WRONG_EFFECT_IN_MMD0, eff);
							quiet = TRUE;
						}

						mednote = (effval |
							((ULONG)(eff & 0x0F) << 8) |
							((ULONG)(note->Inst & 0x0F) << 12) |
							((ULONG)((note->Note ? (note->Note - 12) : 0) & 0x3F) << 16) |
							((ULONG)(note->Inst & 0x20) << 17) |
							((ULONG)(note->Inst & 0x10) << 19)) << 8;

						if (Write (fp, &mednote, MMD0ROW) != MMD0ROW)
							return ERR_READWRITE;
					}
				}
			}

			if (mmd_type)
			{
				if (patt[i].PattName[0])
				{
					/*****************************/
					/* Write BlockInfo structure */
					/*****************************/

					struct BlockInfo blockinfo;

					blockinfoptr += sizeof (blockinfo);

					memset (&blockinfo, 0, sizeof (blockinfo));
					blockinfo.blockname = (UBYTE *)blockinfoptr;
					blockinfo.blocknamelen = strlen (patt[i].PattName) + 1;

					if (Write (fp, &blockinfo, sizeof (blockinfo)) != sizeof (blockinfo))
						return ERR_READWRITE;

					if (Write (fp, patt[i].PattName, blockinfo.blocknamelen) != blockinfo.blocknamelen)
						return ERR_READWRITE;

					blockinfoptr += sizeof (struct BlockInfo) + blockinfo.blocknamelen;

					/* Pad to words */
					if (blockinfoptr & 1)
					{
						UBYTE dummy;

						if (Write (fp, &dummy, 1) != 1)
							return ERR_READWRITE;

						blockinfoptr++;
					}
				}
			}
			else	/* MMD0 */
			{
				/* Add a pad byte when length is not even */
				if ((patt[i].Lines * patt[i].Tracks) & 1)
				{
					UBYTE dummy = 0;

					if (Write (fp, &dummy, 1) != 1)
						return ERR_READWRITE;
				}
			}
		}	/* End for (si->NumPatterns) */
	}


	/**********************/
	/* Write samples data */
	/**********************/
	{
		struct InstrHdr instrhdr;
		UBYTE *samp;
		LONG l;
		BOOL free_samp;

		DisplayAction (MSG_WRITING_INSTDATA);

		for (i = 1; i <= lastinstr; i++)
		{
			free_samp = FALSE;

			l = inst[i].Length;
			if (!l || (!inst[i].SampleData)) continue;

			if (DisplayProgress (i, lastinstr))
				return ERROR_BREAK;

			if (halve_instr)
			{
				if (samp = AllocVec (l, MEMF_ANY))
				{
					/* Halve Volume */
					for (k = 0; k < l; k++)
						samp[k] = inst[i].SampleData[k] >> 1;

					free_samp = TRUE;	/* Free this buffer when you are done */
				}
				else
				{
					ShowMessage (MSG_NO_MEM_TO_HALVE, i);
					samp = inst[i].SampleData;
				}
			}
			else samp = inst[i].SampleData;

			instrhdr.length = l;
			instrhdr.type = SAMPLE;

			if (Write (fp, &instrhdr, sizeof (instrhdr)) != sizeof (instrhdr))
				return ERR_READWRITE;

			if (l)
				if (Write (fp, samp, l) != l)
					return ERR_READWRITE;

			if (free_samp)
				FreeVec (samp);
		}
	}

	return RETURN_OK;
}
